from enum import Enum

import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import pairwise
from pynndescent import NNDescent

from hnne.cool_functions import cool_mean, cool_max_radius
from hnne.point_spreading import norm_angles, norm_angles_3d


class PreliminaryEmbedding(str, Enum):
    pca = 'pca'
    pca_centroids = 'pca_centroids'
    random_linear = 'random_linear'


def project_with_pca_centroids(data, partitions, partition_sizes, dim=2, min_number_of_anchors=1000, verbose=False):
    pca = PCA(n_components=dim)
    large_partitions = np.where(np.array(partition_sizes) > min_number_of_anchors)[0]
    partition_idx = large_partitions.max() if any(large_partitions) else 0
    if verbose:
        print(f'Projecting on the {partition_idx}th partition with {partition_sizes[partition_idx]} anchors.')
    selected_anchors = cool_mean(data, partitions[:, partition_idx])
    pca.fit(selected_anchors)
    return pca.transform(data), pca


def project_with_pca(data, dim=2):
    pca = PCA(n_components=dim)
    transformed_data = pca.fit_transform(data)
    return transformed_data, pca


def project_points(
        data, 
        dim=2, 
        preliminary_embedding='pca',
        partition_sizes=None, 
        partitions=None,
        verbose=False
):
    scaler = StandardScaler()
    data = scaler.fit_transform(data)
    pca = None

    if preliminary_embedding == PreliminaryEmbedding.pca:
        projected_points, pca = project_with_pca(data, dim=dim)
    elif len(data) < dim or preliminary_embedding == PreliminaryEmbedding.pca_centroids:
        projected_points, pca = project_with_pca_centroids(
            data,
            partitions,
            partition_sizes,
            dim=dim,
            verbose=verbose)
    elif preliminary_embedding == PreliminaryEmbedding.random_linear:
        random_components = np.random.random((data.shape[1], dim))
        projected_points = np.dot(data, random_components)
    else:
        raise ValueError(f'Invalid preliminary embedding: {preliminary_embedding}')
        
    # TODO: Handle the case where pca is not defined (e.g. random projection)
    # TODO: Add an option to perform full (randomized) PCA
    return projected_points, pca, scaler


def get_finch_anchors(projected_points, partitions=None):   
    all_projected_anchors = []    
    for i in range(partitions.shape[-1]):
        projected_anchors = cool_mean(projected_points, partitions[:, i])
        all_projected_anchors.append(projected_anchors)
    return all_projected_anchors


def move_projected_points_to_anchors(
        points, 
        anchors, 
        partition,
        radius=.9,
        real_nn_threshold=30000,
        verbose=False
):
    if anchors.shape[0] <= real_nn_threshold:
        distance_matrix = pairwise.pairwise_distances(anchors, anchors, metric='euclidean')
        np.fill_diagonal(distance_matrix, 1e12)
        nearest_neighbor_idx = np.argmin(distance_matrix, axis=1).flatten()
    else:
        if verbose:
            print('Using ann to approximate 1-nns of the projected points...')
        knn_index = NNDescent(
            anchors, 
            n_neighbors=2, 
            metric='euclidean', 
            verbose=verbose)
        nns, _ = knn_index.neighbor_graph
        nearest_neighbor_idx = nns[:, 1]
        
    anchor_distances_from_nns = np.linalg.norm(anchors - anchors[nearest_neighbor_idx], axis=1, keepdims=True)
    anchor_radii = anchor_distances_from_nns * radius
    
    anchors_per_point = anchors[partition]
    anchor_radii_per_point = anchor_radii[partition]
    
    points_mean_per_partition = cool_mean(points, partition)
    points_centered = points - points_mean_per_partition[partition]

    anchors_max_radius = cool_max_radius(points_centered, partition)
    anchors_max_radius = np.where(anchors_max_radius == 0., 1., anchors_max_radius)
    points_max_radius = np.expand_dims(anchors_max_radius[partition], axis=1)

    return (
        anchors_per_point + anchor_radii_per_point * points_centered / points_max_radius,
        anchor_radii[:, 0],
        points_mean_per_partition,
        anchors_max_radius)


def multi_step_projection(
    data, 
    partitions,
    partition_labels,
    radius,
    ann_threshold,
    dim=2,
    partition_sizes=None,
    preliminary_embedding='pca',
    verbose=False
):
    projected_points, pca, scaler = project_points(
        data, 
        dim=dim, 
        preliminary_embedding=preliminary_embedding,
        partition_sizes=partition_sizes, 
        partitions=partitions,
        verbose=verbose
    )

    if verbose:
        print(partition_sizes)
    reversed_partition_range = list(reversed(range(partitions.shape[1])))
    projected_anchors = get_finch_anchors(projected_points, partitions=partitions)
    
    projected_anchors = [projected_points] + projected_anchors
    curr_anchors = projected_anchors[-1]
    
    anchor_radii = []
    moved_anchors = [curr_anchors]
    points_means = []
    points_max_radii = []
    inflation_params_list = []
    for cnt, i in enumerate(reversed_partition_range):
        if i == 0:
            partition_mapping = partitions[:, 0]
        else:
            partition_mapping = partition_labels[i - 1]
        
        current_points = projected_anchors[i]
        if dim == 2:
            thetas = np.linspace(0, np.pi/2, 6)
            current_points, inflation_params = norm_angles(
                current_points,
                thetas,
                partition_mapping)
            inflation_params_list.append(inflation_params)
        if dim == 3:
            alphas, beta, gammas = 3*[np.linspace(0, np.pi/2, 6)]
            current_points, inflation_params = norm_angles_3d(
                current_points,
                alphas,
                beta,
                gammas,
                partition_mapping)
            inflation_params_list.append(inflation_params)

        curr_anchors, radii, points_mean, points_max_radius = move_projected_points_to_anchors(
            current_points,
            curr_anchors, 
            partition_mapping,
            radius=radius,
            real_nn_threshold=ann_threshold,
            verbose=verbose
        )

        anchor_radii.append(radii)
        moved_anchors.append(curr_anchors)
        points_means.append(points_mean)
        points_max_radii.append(points_max_radius)

    return curr_anchors, anchor_radii, moved_anchors, pca, scaler, points_means, points_max_radii, inflation_params_list



####################################################################################################################
# Code adapted with few changes from the FINCH clustering algorithm to generate the levels of the h-NNE hierarchy. #
# FINCH repository: https://github.com/ssarfraz/FINCH-Clustering                                                   #
# Original script: https://github.com/ssarfraz/FINCH-Clustering/blob/master/python/finch.py                        #
####################################################################################################################

import numpy as np
from sklearn import metrics
import scipy.sparse as sp
from pynndescent import NNDescent

from hnne.cool_functions import cool_mean


def clust_rank(
        mat,
        initial_rank=None,
        metric='cosine',
        verbose=False,
        ann_threshold=40000):
    knn_index = None
    s = mat.shape[0]
    if initial_rank is not None:
        orig_dist = []
    elif s <= ann_threshold:
        orig_dist = metrics.pairwise.pairwise_distances(mat, mat, metric=metric)
        np.fill_diagonal(orig_dist, 1e12)
        initial_rank = np.argmin(orig_dist, axis=1)
    else:
        if verbose:
            print('Using PyNNDescent to compute 1st-neighbours at this step ...')
        knn_index = NNDescent(
            mat, 
            n_neighbors=2, 
            metric=metric,
            verbose=verbose)
        result, orig_dist = knn_index.neighbor_graph
        initial_rank = result[:, 1]
        orig_dist[:, 0] = 1e12
        if verbose:
            print('Step PyNNDescent done ...')

    sparce_adjacency_matrix = sp.csr_matrix(
        (np.ones_like(initial_rank, dtype=np.float32),
         (np.arange(0, s), initial_rank)),
        shape=(s, s))
    
    return sparce_adjacency_matrix, orig_dist, initial_rank, knn_index


def get_clust(a, orig_dist, min_sim=None):
    if min_sim is not None:
        a[np.where((orig_dist * a.toarray()) > min_sim)] = 0

    num_clust, u = sp.csgraph.connected_components(csgraph=a, directed=True, connection='weak', return_labels=True)
    return u, num_clust


def get_merge(c, u, data):
    if len(c) != 0:
        _, ig = np.unique(c, return_inverse=True)
        c = u[ig]
    else:
        c = u
    
    mat = cool_mean(data, c)
    
    return c, mat


def update_adj(adj, d):
    # Update adj, keep one merge at a time
    idx = adj.nonzero()
    v = np.argsort(d[idx])
    v = v[:2]
    x = [idx[0][v[0]], idx[0][v[1]]]
    y = [idx[1][v[0]], idx[1][v[1]]]
    a = sp.lil_matrix(adj.get_shape())
    a[x, y] = 1
    return a


def req_numclust(c, data, req_clust, distance):
    iter_ = len(np.unique(c)) - req_clust
    c_, mat = get_merge([], c, data)
    for i in range(iter_):
        adj, orig_dist, _, _ = clust_rank(mat, initial_rank=None, metric=distance)
        adj = update_adj(adj, orig_dist)
        u, _ = get_clust(adj, [], min_sim=None)
        c_, mat = get_merge(c_, u, data)
    return c_


# noinspection PyPep8Naming
def FINCH(
        data,
        initial_rank=None,
        distance='cosine',
        ensure_early_exit=True,
        verbose=True,
        ann_threshold=40000):
    """FINCH clustering algorithm.

    Parameters
    ----------
        data: array, shape (n_samples, n_features)
            Input matrix with features in rows.

        initial_rank: array, shape (n_samples, 1) (optional)
            First integer neighbor indices.

        distance: str (default 'cosine')
            One of ['cityblock', 'cosine', 'euclidean', 'l1', 'l2', 'manhattan'] Recommended 'cosine'.

        ensure_early_exit: bool (default True)
            May help in large, high dim datasets, ensure purity of merges and helps early exit.

        verbose: bool (default True)
            Print verbose output.

        ann_threshold: int (default 40000)
            Data size threshold below which nearest neighbors are approximated with ANNs.

    Returns
    -------
        c: array of shape (n_samples, n_partitions)
            Matrix with labels indicating cluster participation. There is one column per partition.

        num_clust: array of shape (n_partitions)
            Number of clusters per partition.

        partition_clustering: list of arrays of shapes equal to the values of num_clust
            List of arrays with labels indicating the centroids cluster participation per level.

        lowest_level_centroids: array of shape (num_clust[0], n_features)
            The feature coordinates of the lowest level centroids.

    References
    ----------
        The code implements the FINCH algorithm described in our CVPR 2019 paper
        [1] Sarfraz et al. "Efficient Parameter-free Clustering Using First Neighbor Relations", CVPR2019
             https://arxiv.org/abs/1902.11266
        For academic purpose only. The code or its re-implementation should not be used for commercial use.
        Please contact the author below for licensing information.
        Copyright
        M. Saquib Sarfraz (saquib.sarfraz@kit.edu)
        Karlsruhe Institute of Technology (KIT)
    """
    data = data.astype(np.float32)

    min_sim = None
    
    adj, orig_dist, first_neighbors, _ = clust_rank(
        data,
        initial_rank,
        distance,
        verbose=verbose,
        ann_threshold=ann_threshold
    )
    initial_rank = None
    
    group, num_clust = get_clust(adj, [], min_sim)
    
    c, mat = get_merge([], group, data)
    lowest_level_centroids = mat
    
    if verbose:
        print('Level 0: {} clusters'.format(num_clust))

    if ensure_early_exit:
        if orig_dist.shape[-1] > 2:
            min_sim = np.max(orig_dist * adj.toarray())

    exit_clust = 2
    c_ = c
    k = 1
    num_clust = [num_clust]
    partition_clustering = []
    while exit_clust > 1:    
        adj, orig_dist, first_neighbors, knn_index = clust_rank(
            mat,
            initial_rank,
            distance,
            verbose=verbose,
            ann_threshold=ann_threshold
        )

        u, num_clust_curr = get_clust(adj, orig_dist, min_sim)

        partition_clustering.append(u)
        
        c_, mat = get_merge(c_, u, data)
        c = np.column_stack((c, c_))
        
        num_clust.append(num_clust_curr)
        exit_clust = num_clust[-2] - num_clust_curr

        if num_clust_curr == 1 or exit_clust <= 1:
            num_clust = num_clust[:-1]
            c = c[:, :-1]
            break

        if verbose:
            print('Level {}: {} clusters'.format(k, num_clust[k]))
        k += 1

    return c, num_clust, partition_clustering, lowest_level_centroids



import pickle
from dataclasses import dataclass
from typing import Optional, List, Any

import numpy as np
from sklearn import metrics
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler
from sklearn.base import BaseEstimator
from pynndescent import NNDescent

from hnne.finch_clustering import FINCH
from hnne.hierarchical_projection import multi_step_projection, PreliminaryEmbedding


@dataclass
class HierarchyParameters:
    partitions: np.ndarray
    partition_sizes: np.ndarray
    partition_labels: np.ndarray
    lowest_level_centroids: np.ndarray


@dataclass
class ProjectionParameters:
    pca: Optional[PCA]
    scaler: StandardScaler
    projected_centroid_radii: List[np.ndarray]
    projected_centroids: List[np.ndarray]
    points_means: List[np.ndarray]
    points_max_radii: List[np.ndarray]
    inflation_params_list: List[Any]
    knn_index_transform: Optional[Any]


class HNNE(BaseEstimator):
    """Hierarchical 1-Nearest Neighbor graph based Embedding

    A fast hierarchical dimensionality reduction algorithm.

    Parameters
    ----------
    dim: int (default 2)
        The dimension of the target space of the projection.

    metric: str (default 'cosine')
        The metric used to compute the distances when forming the h-nne hierarchy levels. Its value should be supported
        by both sklearn and pynndescent. Some possible values: 'cityblock', 'cosine', 'euclidean',
        'l1', 'l2', 'manhattan'.

    radius: float (default 0.45)
        The radius used to place points around centroids as a portion of the distance between nearest neighbor anchors.
        Though the theoretical value which guarantees no overlaps between anchor points is 0.2, 0.45 is a value which
        provides in practice denser visualizations with minimal loss in performance.

    ann_threshold: int (default 40000)
        A threshold above which approximate nearest neighbors will be computed instead of real nearest neighbors when
        building the levels of h-nne.

    preliminary_embedding: str (default 'pca')
        The preliminary embedding used to initiate h-nne. In terms of performance pca > pca_centroids > random_linear
        and in terms of speed performance pca < pca_centroids < random_linear.

    Attributes
    ----------
    min_size_top_level: int (default 3)
        The minimum number of centroids existing on the top level of the hierarchy. To achieve this minimum, the top
        levels which have fewer centroids are removed.

    hierarchy_parameters: Optional[HierarchyParameters]
        An object holding the parameters which encode the h-nne hierarchy. They are saved during fitting and can be
        reused both during projecting new points or projecting again with different parameters, e.g. dim.
    """
    def __init__(
            self,
            dim: int = 2,
            metric: str = 'cosine',
            radius: float = 0.4,
            ann_threshold: int = 40000,
            preliminary_embedding: str = 'pca'
    ):
        self.dim = dim
        self.radius = radius
        self.ann_threshold = ann_threshold
        try:
            preliminary_embedding = PreliminaryEmbedding[preliminary_embedding]
        except KeyError:
            raise ValueError(
                f'Invalid preliminary embedding: {preliminary_embedding}. '
                f'Please select one from: {", ".join(PreliminaryEmbedding)}.')
        self.preliminary_embedding = preliminary_embedding
        self.metric = metric
        self.min_size_top_level: int = 3
        self.hierarchy_parameters: Optional[HierarchyParameters] = None
        self.projection_parameters: Optional[ProjectionParameters] = None

    def fit_only_hierarchy(self, X: np.ndarray, verbose: bool = False):
        if verbose:
            print('Building h-NNE hierarchy using FINCH...')
        [
            partitions,
            partition_sizes,
            partition_labels,
            lowest_level_centroids
        ] = FINCH(
            data=X,
            ensure_early_exit=False,
            verbose=verbose,
            distance=self.metric,
            ann_threshold=self.ann_threshold
        )

        large_enough_partitions = np.argwhere(np.array(partition_sizes) >= self.min_size_top_level)
        if len(large_enough_partitions) == 0:
            raise ValueError(
                f'The dataset has too few points resulting to a hierarchy with sizes {partition_sizes}. Please provide'
                f' a larger amount of data till there exists one partition of size {self.min_size_top_level}.')
        max_partition_idx = int(large_enough_partitions.max()) + 1
        if max_partition_idx < len(partition_sizes) and verbose:
            print(f'Removing {len(partition_sizes) - max_partition_idx} levels from the top to start with a level'
                  f'of size at least {self.min_size_top_level}.')
        partition_sizes = partition_sizes[:max_partition_idx]
        partitions = partitions[:, :max_partition_idx]
        partition_labels = partition_labels[:max_partition_idx]

        self.hierarchy_parameters = HierarchyParameters(
            partitions,
            partition_sizes,
            partition_labels,
            lowest_level_centroids
        )

        return partitions, partition_sizes, partition_labels

    # noinspection PyUnusedLocal
    def fit(
            self,
            X: np.ndarray,
            y: np.ndarray = None,
            dim: int = 2,
            verbose: bool = False,
            skip_hierarchy_building_if_done: bool = True
    ):
        """
        Build an h-nne hierarchy based on X and use it to project X.

        Parameters
        ----------
        X: array, shape (n_samples, n_features)
            The data to project.

        y: array, shape (n_samples, )
            Ignored.

        dim: int (default 2)
            Argument used to overwrite the original dimension of the target space of the projection.

        verbose: bool
            If true, plot info and progress messages.

        skip_hierarchy_building_if_done:
            If true, the h-nne hierarchy will be built only on the first run of fit. Warning: if you need to project
            a new dataset with the same HNNE object, then you have to set this to false.
        """
        if self.hierarchy_parameters is not None and skip_hierarchy_building_if_done:
            if verbose:
                print('Skipping the hierarchy construction as it is already available.')
            hparams = self.hierarchy_parameters
            partitions, partition_sizes, partition_labels = \
                hparams.partitions, hparams.partition_sizes, hparams.partition_labels

        else:
            [
                partitions,
                partition_sizes,
                partition_labels
            ] = self.fit_only_hierarchy(X, verbose=verbose)

        if dim is not None and dim != self.dim:
            if verbose:
                print(f'Overwriting the dimensions {self.dim} to the new value {dim}.')
            self.dim = dim

        if verbose:
            print(f'Projecting to {dim} dimensions...')
        [
            projection,
            projected_centroid_radii,
            projected_centroids,
            pca,
            scaler,
            points_means,
            points_max_radii,
            inflation_params_list
        ] = multi_step_projection(
            data=X,
            partitions=partitions,
            partition_labels=partition_labels,
            radius=self.radius,
            ann_threshold=self.ann_threshold,
            dim=self.dim,
            partition_sizes=partition_sizes,
            preliminary_embedding=self.preliminary_embedding,
            verbose=verbose
        ) 

        self.projection_parameters = ProjectionParameters(
            pca=pca,
            scaler=scaler,
            projected_centroid_radii=projected_centroid_radii,
            projected_centroids=projected_centroids,
            points_means=points_means,
            points_max_radii=points_max_radii,
            inflation_params_list=inflation_params_list,
            knn_index_transform=None
        )
        
        return projection
        
    def transform(self, X: np.ndarray, ann_point_combination_threshold: int = 400e6, verbose: bool = False):
        if self.hierarchy_parameters is None or self.projection_parameters is None:
            raise ValueError('Unable to project as h-nne has not been fitted on a dataset.')
        hparams = self.hierarchy_parameters
        pparams = self.projection_parameters

        if verbose:
            print('Finding nearest centroids to new data...')
        if len(hparams.lowest_level_centroids) * len(X) > ann_point_combination_threshold:
            nns = 30
            if pparams.knn_index_transform is None:
                if verbose:
                    print('Setting up once a knn index for the last level centroids...')
                knn_index = NNDescent(
                    hparams.lowest_level_centroids,
                    n_neighbors=nns,
                    metric=self.metric,
                    verbose=verbose)
                knn_index.prepare()
                pparams.knn_index_transform = knn_index
            else:
                knn_index = pparams.knn_index_transform
            nearest_anchor_idxs = knn_index.query(X, k=nns)[0][:, 0]
        else:
            orig_dist = metrics.pairwise.pairwise_distances(X, hparams.lowest_level_centroids, metric=self.metric)
            nearest_anchor_idxs = np.argmin(orig_dist, axis=1)

        if verbose:
            print('Projecting data...')
        # Project the points with pca
        X = pparams.scaler.transform(X)
        X = pparams.pca.transform(X)

        # Apply inflation to points, if applicable
        if self.dim <= 3:
            for rot, norm1_params, norm2_params in pparams.inflation_params_list[-1]:
                m1, s1 = norm1_params
                m1, s1 = m1[nearest_anchor_idxs], s1[nearest_anchor_idxs]
                m2, s2 = norm2_params
                m2, s2 = m2[nearest_anchor_idxs], s2[nearest_anchor_idxs]
                X = (X - m1) / s1
                X = np.dot(X, rot)
                X = (X - m2) / s2
                X = np.dot(X, np.linalg.inv(rot))
                data_norms = np.linalg.norm(X, axis=-1)
                X = np.where(
                    np.expand_dims(data_norms > 1, axis=-1),
                    X / np.expand_dims(data_norms, axis=-1),
                    X)

        # Compute parameters related to the nearest anchors
        projected_nearest_anchors = pparams.projected_centroids[-2][nearest_anchor_idxs]
        max_radii = np.expand_dims(pparams.points_max_radii[-1][nearest_anchor_idxs], axis=-1)
        centroid_radii = np.expand_dims(pparams.projected_centroid_radii[-1][nearest_anchor_idxs], axis=-1)

        # Normalize relative to the maximum anchor group original point radius
        points_mean = pparams.points_means[-1][nearest_anchor_idxs]
        normalized_points = (X - points_mean) / max_radii

        # Scale based on the nearest anchor radii
        return projected_nearest_anchors + normalized_points * centroid_radii

    fit_transform = fit

    def save(self, path):
        with open(path, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
            
    @staticmethod
    def load(path):
        with open(path, 'rb') as f:
            return pickle.load(f)
